chiark / gitweb /
[PATCH] fix stupid off-by-one bug that caused udevstart to die on x86-64 boxes.
[elogind.git] / namedev.c
index 025a14aea4a5ae832c6d9de1987562a859b8112b..52417e3793b655fcca041ff1ead58e7a0ba8d4ce 100644 (file)
--- a/namedev.c
+++ b/namedev.c
 #include <time.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
+#include <sys/sysinfo.h>
 
 #include "libsysfs/sysfs/libsysfs.h"
 #include "list.h"
 #include "udev.h"
+#include "udev_lib.h"
 #include "udev_version.h"
 #include "logging.h"
 #include "namedev.h"
@@ -195,23 +197,21 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
        struct sysfs_attribute *tmpattr;
 
        pos = string;
-
        while (1) {
-               pos = strchr(string, '%');
-               if (pos != NULL) {
-                       pos[0] = '\0';
-                       tail = pos+1;
-                       len = get_format_len(&tail);
-                       c = tail[0];
-                       strfieldcpy(temp, tail+1);
-                       tail = temp;
-               } else {
+               pos = strchr(pos, '%');
+               if (pos == NULL)
                        break;
-               }
-               dbg("format=%c, string='%s', tail='%s'",c , string, tail);
 
+               pos[0] = '\0';
+               tail = pos+1;
+               len = get_format_len(&tail);
+               c = tail[0];
+               strfieldcpy(temp, tail+1);
+               tail = temp;
+               dbg("format=%c, string='%s', tail='%s'",c , string, tail);
                attr = get_format_attribute(&tail);
 
+
                switch (c) {
                case 'b':
                        if (strlen(udev->bus_id) == 0)
@@ -282,6 +282,7 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                        break;
                case '%':
                        strfieldcatmax(string, "%", maxsize);
+                       pos++;
                        break;
                default:
                        dbg("unknown substitution type '%%%c'", c);
@@ -453,7 +454,7 @@ static int execute_program(char *path, char *value, int len)
                        retval = -1;
                }
 
-               if (i > 0 && value[i] == '\n')
+               if (i > 0 && value[i-1] == '\n')
                        i--;
                value[i] = '\0';
                dbg("result is '%s'", value);
@@ -601,6 +602,42 @@ static int match_place(struct config_device *dev, struct sysfs_class_device *cla
        return 0;
 }
 
+static int whitelist_search(struct sysfs_class_device *class_dev)
+{
+       char *sysblock = "/sys/block";
+       int i;
+
+       static char *list[] = {
+               "nb",
+               "ram",
+               "loop",
+               "fd",
+               "md",
+               "dos_cd",
+               "double",
+               "flash",
+               "msd",
+               "rflash",
+               "rom",
+               "rrom",
+               "sbpcd",
+               "pcd",
+               "pf",
+               "scd",
+               "ubd",
+               NULL,
+       };
+
+       if (strncmp(class_dev->path, sysblock, strlen(sysblock)))
+               return 0;
+
+       for (i=0; list[i] != NULL; i++)
+               if (!strncmp(class_dev->name, list[i], strlen(list[i])))
+                       return 1;
+
+       return 0;
+}
+
 static struct sysfs_device *get_sysfs_device(struct sysfs_class_device *class_dev)
 {
        struct sysfs_device *sysfs_device;
@@ -614,37 +651,39 @@ static struct sysfs_device *get_sysfs_device(struct sysfs_class_device *class_de
         * all partitions have the symlink in its parent directory.
         * But we need to watch out for block devices that do not have parents, yet
         * look like a partition (fd0, loop0, etc.)  They all do not have a device
* symlink yet.  We do sit and spin on waiting for them right now, we should
* possibly have a whitelist for these devices here...
         */
        class_dev_parent = sysfs_get_classdev_parent(class_dev);
-       if (class_dev_parent) 
-               dbg("Really a partition");
+       if (class_dev_parent != NULL
+               dbg("given class device has a parent, use this instead");
 
        tspec.tv_sec = 0;
        tspec.tv_nsec = 10000000;  /* sleep 10 millisec */
        loop = 10;
        while (loop--) {
-               if (udev_sleep)
+               if (udev_sleep) {
+                       if (whitelist_search(class_dev)) {
+                               sysfs_device = NULL;
+                               goto exit;
+                       }
                        nanosleep(&tspec, NULL);
+               }
+
                if (class_dev_parent)
                        sysfs_device = sysfs_get_classdev_device(class_dev_parent);
                else
                        sysfs_device = sysfs_get_classdev_device(class_dev);
-
                if (sysfs_device != NULL)
                        goto device_found;
        }
        dbg("timed out waiting for device symlink, continuing on anyway...");
-       
+
 device_found:
         /* We have another issue with just the wait above - the sysfs part of
         * the kernel may not be quick enough to have created the link to the
         * device under the "bus" subsystem. Due to this, the sysfs_device->bus
         * will not contain the actual bus name :(
-        *
-        * Libsysfs now provides a new API sysfs_get_device_bus(), so use it
-        * if needed
         */
         if (sysfs_device) {
                if (sysfs_device->bus[0] != '\0')
@@ -778,6 +817,7 @@ int namedev_name_device(struct sysfs_class_device *class_dev, struct udevice *ud
        struct sysfs_device *sysfs_device = NULL;
        struct config_device *dev;
        struct perm_device *perm;
+       struct sysinfo info;
        char *pos;
 
        udev->mode = 0;
@@ -815,14 +855,23 @@ int namedev_name_device(struct sysfs_class_device *class_dev, struct udevice *ud
                        }
 
                        if (dev->symlink[0] != '\0') {
+                               char temp[NAME_SIZE];
+
                                info("configured rule in '%s' at line %i applied, added symlink '%s'",
                                     dev->config_file, dev->config_line, dev->symlink);
+                               strfieldcpy(temp, dev->symlink);
+                               apply_format(udev, temp, sizeof(temp), class_dev, sysfs_device);
                                if (udev->symlink[0] != '\0')
                                        strfieldcat(udev->symlink, " ");
-                               strfieldcat(udev->symlink, dev->symlink);
+                               strfieldcat(udev->symlink, temp);
                        }
 
                        if (dev->name[0] != '\0') {
+                               /* apply all_partitions flag only at a main block device */
+                               if (dev->partitions > 0 &&
+                                   (udev->type != 'b' || udev->kernel_number[0] != '\0'))
+                                       continue;
+
                                info("configured rule in '%s' at line %i applied, '%s' becomes '%s'",
                                     dev->config_file, dev->config_line, udev->kernel_name, dev->name);
                                strfieldcpy(udev->name, dev->name);
@@ -830,22 +879,29 @@ int namedev_name_device(struct sysfs_class_device *class_dev, struct udevice *ud
                        }
                }
        }
-
        /* no rule was found so we use the kernel name */
        strfieldcpy(udev->name, udev->kernel_name);
-       goto done;
+       if (udev->type == 'n')
+               goto done;
+       else
+               goto perms;
 
 found:
        apply_format(udev, udev->name, sizeof(udev->name), class_dev, sysfs_device);
-       apply_format(udev, udev->symlink, sizeof(udev->symlink), class_dev, sysfs_device);
+       strfieldcpy(udev->config_file, dev->config_file);
+       udev->config_line = dev->config_line;
+
+       if (udev->type == 'n')
+               goto done;
+
        udev->partitions = dev->partitions;
 
-done:
        /* get permissions given in rule */
        set_empty_perms(udev, dev->mode,
                              dev->owner,
                              dev->group);
 
+perms:
        /* get permissions given in config file or set defaults */
        perm = find_perm(udev->name);
        if (perm != NULL) {
@@ -861,6 +917,11 @@ done:
        dbg("name, '%s' is going to have owner='%s', group='%s', mode = %#o",
            udev->name, udev->owner, udev->group, udev->mode);
 
+done:
+       /* store time of action */
+       sysinfo(&info);
+       udev->config_uptime = info.uptime;
+
        return 0;
 }
@@ -614,37 +651,39 @@ static struct sysfs_device *get_sysfs_device(struct sysfs_class_device *class_de
         * But we need to watch out for block devices that do not have parents, yet
         * look like a partition (fd0, loop0, etc.)  They all do not have a device
-        * symlink yet.  We do sit and spin on waiting for them right now, we should
-        * possibly have a whitelist for these devices here...
+        * symlink yet.  We do sit and spin on waiting for them right now unless
+        * they happen to be in the whitelist in which case we exit.
         */
        class_dev_parent = sysfs_get_classdev_parent(class_dev);